from requests import Response


class MockResponse(Response):
    def __init__(self, response_object, status_code: int):
        self.status_code = status_code
        self.response_object = response_object

    def json(self):
        return self.response_object



import os
import json

from arthurai.core.attributes import AttributeCategory, AttributeBin, ArthurAttribute
from arthurai.core.models import ArthurModel, ExplainabilityParameters
from arthurai.core.alerts import AlertRule, AlertRuleSeverity,AlertRuleBound
from arthurai.common.constants import ValueType, InputType, Stage, OutputType
from tests.test_request_models import load_json_string, load_json

# *************************************************************
# fixtures for the TestExplainabilityParametersObject class
# *************************************************************
explainability_response_objects = {
    1: ExplainabilityParameters(explanation_algo='lime',
                                enabled=True,
                                model_server_cpu='800m',
                                model_server_memory='1Gi',
                                explanation_nsamples=4000),
    2: ExplainabilityParameters(explanation_algo='lime',
                                enabled=True),
    3: ExplainabilityParameters(explanation_algo='lime',
                                enabled=True,
                                model_server_cpu='800m',
                                explanation_nsamples=4000)
}

explainability_json_strings = {
    1: '{"enabled": true, "explanation_algo": "lime", "model_server_cpu": "800m", "model_server_memory": "1Gi", '
       '"explanation_nsamples": 4000}',
    2: '{"enabled": true, "explanation_algo": "lime"}',
    3: '{"enabled": true, "explanation_algo": "lime", "model_server_cpu": "800m", "explanation_nsamples": 4000, '
       '"foo": "bar"}'
}

explainability_dicts = {
    1: {
        "explanation_algo": "lime",
        "enabled": True,
        "model_server_cpu": "800m",
        "model_server_memory": "1Gi",
        "explanation_nsamples": 4000
    },
    2: {
        "explanation_algo": "lime",
        "enabled": True
    },
    3: {
        "explanation_algo": "lime",
        "enabled": True,
        "model_server_cpu": "800m",
        "explanation_nsamples": 4000
    }
}

# *************************************************************
# fixtures for the TestExplainabilityParametersObject class
# *************************************************************
categories_1 = [
    AttributeCategory(value='value1'),
    AttributeCategory(value='value2'),
    AttributeCategory(value='value3')
]

categories_2 = [
    AttributeCategory(label='positive_label', value='1'),
    AttributeCategory(label='negative_label', value='0')
]

bins_1 = [
    AttributeBin(continuous_start=0, continuous_end=5),
    AttributeBin(continuous_start=5, continuous_end=18),
    AttributeBin(continuous_start=18, continuous_end=30),
    AttributeBin(continuous_start=30, continuous_end=65),
    AttributeBin(continuous_start=65, continuous_end=100)
]

model_attribute_objects = {
    1: ArthurAttribute(id='8514e278-24d1-4e20-b209-3c77eb2b247f',
                       name='attr_1',
                       value_type=ValueType.Integer,
                       stage=Stage.ModelPipelineInput,
                       categorical=True,
                       monitor_for_bias=False,
                       categories=categories_1,
                       is_unique=False,
                       is_positive_predicted_attribute=False),
    2: ArthurAttribute(id='8514e278-24d1-4e20-b209-3c77eb2b247f',
                       name='attr_2',
                       position=1,
                       value_type=ValueType.Boolean,
                       stage=Stage.ModelPipelineInput,
                       categorical=True,
                       monitor_for_bias=True,
                       categories=categories_2,
                       is_unique=False,
                       is_positive_predicted_attribute=True),
    3: ArthurAttribute(id='04418f36-529b-451e-8524-c57d09d9b02f',
                       name='attr_3',
                       value_type=ValueType.Integer,
                       stage=Stage.ModelPipelineInput,
                       categorical=False,
                       monitor_for_bias=False,
                       min_range=0,
                       max_range=100,
                       bins=bins_1,
                       is_unique=False,
                       is_positive_predicted_attribute=False),
    4: ArthurAttribute(id='8514e278-24d1-4e20-b209-3c77eb2b247f',
                       name='attr_3',
                       value_type=ValueType.Integer,
                       stage=Stage.ModelPipelineInput,
                       categorical=False,
                       monitor_for_bias=False,
                       min_range=0,
                       max_range=100,
                       bins=bins_1,
                       is_unique=False,
                       is_positive_predicted_attribute=False)
}

# load_json()

model_attribute_json_strings = {
    1: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'attribute_with_categories.json')),
    2: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'attribute_with_categories_and_labels.json')),
    3: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'attribute_with_bins.json')),
    4: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'attribute_with_unknown_properties.json'))
}

model_attribute_dicts = {
    1: {
        "id": "8514e278-24d1-4e20-b209-3c77eb2b247f",
        "name": "attr_1",
        "value_type": ValueType.Integer,
        "stage": Stage.ModelPipelineInput,
        "categorical": True,
        "monitor_for_bias": False,
        "categories": [
            {
                "value": "value1"
            },
            {
                "value": "value2"
            },
            {
                "value": "value3"
            }
        ],
        "is_unique": False,
        "is_positive_predicted_attribute": False
    },
    2: {
        "id": "8514e278-24d1-4e20-b209-3c77eb2b247f",
        "name": "attr_2",
        "value_type": ValueType.Boolean,
        "stage": Stage.ModelPipelineInput,
        "position": 1,
        "categorical": True,
        "monitor_for_bias": True,
        "categories": [
            {
                "value": "1",
                "label": "positive_label"
            },
            {
                "value": "0",
                "label": "negative_label"
            }
        ],
        "is_unique": False,
        "is_positive_predicted_attribute": True
    },
    3: {
        "id": "04418f36-529b-451e-8524-c57d09d9b02f",
        "name": "attr_3",
        "value_type": ValueType.Integer,
        "stage": Stage.ModelPipelineInput,
        "categorical": False,
        "min_range": 0,
        "max_range": 100,
        "monitor_for_bias": False,
        "bins": [
            {
                "continuous_start": 0,
                "continuous_end": 5
            },
            {
                "continuous_start": 5,
                "continuous_end": 18
            },
            {
                "continuous_start": 18,
                "continuous_end": 30
            },
            {
                "continuous_start": 30,
                "continuous_end": 65
            },
            {
                "continuous_start": 65,
                "continuous_end": 100
            }
        ],
        "is_unique": False,
        "is_positive_predicted_attribute": False
    }
}

# *************************************************************
# fixtures for the TestArthurModelObject class
# *************************************************************

model_response_objects = {
    1: ArthurModel(id='ac55c7b4-2db7-4902-8cc3-969ed67a20c8',
                   partner_model_id='3456',
                   input_type=InputType.Tabular,
                   output_type=OutputType.Regression,
                   explainability=explainability_response_objects[1],
                   display_name='test model 1',
                   archived=False,
                   created_at='2020-05-15T15:55:17.631797',
                   updated_at='2020-05-15T16:01:07.961584',
                   attributes=[model_attribute_objects[1], model_attribute_objects[2]],
                   tags=['keyword1', 'keyword2'],
                   classifier_threshold=0.5,
                   client=None),
    2: ArthurModel(id='ac55c7b4-2db7-4902-8cc3-969ed67a20c8',
                   partner_model_id='3456',
                   input_type=InputType.Tabular,
                   output_type=OutputType.Regression,
                   explainability=explainability_response_objects[2],
                   archived=False,
                   created_at='2020-05-15T15:55:17.631797',
                   updated_at='2020-05-15T16:01:07.961584',
                   attributes=[model_attribute_objects[3]],
                   classifier_threshold=0.5,
                   client=None),
    3: ArthurModel(id='ac55c7b4-2db7-4902-8cc3-969ed67a20c8',
                   partner_model_id='3456',
                   input_type=InputType.Tabular,
                   output_type=OutputType.Regression,
                   explainability=explainability_response_objects[2],
                   archived=False,
                   created_at='2020-05-15T15:55:17.631797',
                   updated_at='2020-05-15T16:01:07.961584',
                   attributes=[model_attribute_objects[3]],
                   classifier_threshold=0.5,
                   client=None)
}

model_response_json_strings = {
    1: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'model_with_categories.json')),
    2: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'model_with_attributes.json')),
    3: load_json_string(os.path.join(os.path.dirname(__file__), 'resources', 'model_with_unknown_properties.json'))
}

model_response_dicts = {
    1: {
        "partner_model_id": "3456",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Regression,
        "explainability": {
            "explanation_algo": "lime",
            "enabled": True,
            "model_server_cpu": "800m",
            "model_server_memory": "1Gi",
            "explanation_nsamples": 4000
        },
        "id": "ac55c7b4-2db7-4902-8cc3-969ed67a20c8",
        "display_name": "test model 1",
        "is_batch": False,
        "archived": False,
        "created_at": "2020-05-15T15:55:17.631797",
        "updated_at": "2020-05-15T16:01:07.961584",
        "attributes": [
            {
                "id": "8514e278-24d1-4e20-b209-3c77eb2b247f",
                "name": "attr_1",
                "value_type": ValueType.Integer,
                "stage": Stage.ModelPipelineInput,
                "categorical": True,
                "monitor_for_bias": False,
                "categories": [
                    {
                        "value": "value1"
                    },
                    {
                        "value": "value2"
                    },
                    {
                        "value": "value3"
                    }
                ],
                "is_unique": False,
                "is_positive_predicted_attribute": False
            },
            {
                "id": "8514e278-24d1-4e20-b209-3c77eb2b247f",
                "name": "attr_2",
                "value_type": ValueType.Boolean,
                "stage": Stage.ModelPipelineInput,
                "position": 1,
                "categorical": True,
                "monitor_for_bias": True,
                "categories": [
                    {
                        "value": "1",
                        "label": "positive_label"
                    },
                    {
                        "value": "0",
                        "label": "negative_label"
                    }
                ],
                "is_unique": False,
                "is_positive_predicted_attribute": True
            }
        ],
        "tags": [
            "keyword1",
            "keyword2"
        ],
        "classifier_threshold": 0.5
    },
    2: {
        "partner_model_id": "3456",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Regression,
        "explainability": {
            "explanation_algo": "lime",
            "enabled": True
        },
        "id": "ac55c7b4-2db7-4902-8cc3-969ed67a20c8",
        "is_batch": False,
        "archived": False,
        "created_at": "2020-05-15T15:55:17.631797",
        "updated_at": "2020-05-15T16:01:07.961584",
        "attributes": [
            {
                "id": "04418f36-529b-451e-8524-c57d09d9b02f",
                "name": "attr_3",
                "value_type": ValueType.Integer,
                "stage": Stage.ModelPipelineInput,
                "categorical": False,
                "min_range": 0,
                "max_range": 100,
                "monitor_for_bias": False,
                "bins": [
                    {
                        "continuous_start": 0,
                        "continuous_end": 5
                    },
                    {
                        "continuous_start": 5,
                        "continuous_end": 18
                    },
                    {
                        "continuous_start": 18,
                        "continuous_end": 30
                    },
                    {
                        "continuous_start": 30,
                        "continuous_end": 65
                    },
                    {
                        "continuous_start": 65,
                        "continuous_end": 100
                    }
                ],
                "is_unique": False,
                "is_positive_predicted_attribute": False
            }
        ],
        "classifier_threshold": 0.5
    },
    3: {
        "partner_model_id": "3456",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Regression,
        "explainability": {
            "explanation_algo": "lime",
            "enabled": True
        },
        "id": "ac55c7b4-2db7-4902-8cc3-969ed67a20c8",
        "is_batch": False,
        "archived": False,
        "created_at": "2020-05-15T15:55:17.631797",
        "updated_at": "2020-05-15T16:01:07.961584",
        "attributes": [
            {
                "id": "04418f36-529b-451e-8524-c57d09d9b02f",
                "name": "attr_3",
                "value_type": ValueType.Integer,
                "stage": Stage.ModelPipelineInput,
                "categorical": False,
                "min_range": 0,
                "max_range": 100,
                "monitor_for_bias": False,
                "bins": [
                    {
                        "continuous_start": 0,
                        "continuous_end": 5
                    },
                    {
                        "continuous_start": 5,
                        "continuous_end": 18
                    },
                    {
                        "continuous_start": 18,
                        "continuous_end": 30
                    },
                    {
                        "continuous_start": 30,
                        "continuous_end": 65
                    },
                    {
                        "continuous_start": 65,
                        "continuous_end": 100
                    }
                ],
                "is_unique": False,
                "is_positive_predicted_attribute": False
            }
        ],
        "classifier_threshold": 0.5
    }
}


# *************************************************************
# fixtures for AlertRules
# *************************************************************
AlertRuleObject_1 = AlertRule(
    bound=AlertRuleBound.Upper,
    threshold=0.5,
    metric_id="e14e6aac-0c94-4a78-a104-871f70b8b476",
    severity=AlertRuleSeverity.Warning
)

AlertRuleDict_1 = {
    "bound": "upper",
    "threshold": 0.5,
    "metric_id": "e14e6aac-0c94-4a78-a104-871f70b8b476",
    "severity": "warning",
    "enabled": True
}

AlertRuleJson_1 = json.dumps(load_json(os.path.join(os.path.dirname(__file__), 'resources', 'alert_rule_1.json')))

AlertRuleObject_2 = AlertRule(
    bound=AlertRuleBound.Upper,
    threshold=0.5,
    metric_id="e14e6aac-0c94-4a78-a104-871f70b8b476",
    severity=AlertRuleSeverity.Warning
)

AlertRuleDict_2 = {
    "bound": "upper",
    "threshold": 0.5,
    "metric_id": "e14e6aac-0c94-4a78-a104-871f70b8b476",
    "severity": "warning",
    "enabled": True,
}

AlertRuleJson_2 = json.dumps(load_json(os.path.join(os.path.dirname(__file__), 'resources', 'alert_rule_2.json')))

AlertRuleObject_3 = AlertRule(
    bound=AlertRuleBound.Upper,
    threshold=0.5,
    metric_id="e14e6aac-0c94-4a78-a104-871f70b8b476",
    severity=AlertRuleSeverity.Warning,
    enabled=False,
    id="37c8a2cb-b607-4b77-9a89-9d34ee6e3516"
)

AlertRuleDict_3 = {
    "bound": "upper",
    "threshold": 0.5,
    "metric_id": "e14e6aac-0c94-4a78-a104-871f70b8b476",
    "severity": "warning",
    "enabled": False,
    "id": "37c8a2cb-b607-4b77-9a89-9d34ee6e3516",
}

AlertRuleJson_3 = json.dumps(load_json(os.path.join(os.path.dirname(__file__), 'resources', 'alert_rule_3.json')))




import pandas as pd

from arthurai.common.constants import ValueType
from arthurai.core.models import ArthurModel


class TestModelsServiceObject:

    def test_get_attribute_data(self):
        """
        Currently, tests that float attributes that only take on values of
        0.0 or 1.0 are converted to categorical.
        """

        value_type = ValueType.Float

        df = pd.DataFrame({
            'float_binary': [0.0, 1.0, 1.0, 1.0, 0.0],
            'true_float': [0.0, 0.2, 0.4, 0.6, 0.8]
        })

        binary_attr = ArthurModel._get_attribute_data(value_type, df.float_binary)
        float_attr = ArthurModel._get_attribute_data(value_type, df.true_float)

        assert binary_attr["categorical"] is True
        assert float_attr["categorical"] is False

    def test_unstructured_text_get_attribute_data(self):
        """
        Ensure non-unique unstructured text attributes don't get
        categories
        """
        value_type = ValueType.Unstructured_Text
        df = pd.DataFrame({
            "text": ["test", "test", "more", "things"],
            "target": [0,1,2,3]
        })

        text_attr = ArthurModel._get_attribute_data(value_type, df.text)

        assert text_attr['is_unique'] == False
        assert text_attr['categorical'] == True
        assert 'categories' not in text_attr



from arthurai.core.attributes import AttributeCategory


class TestAttributeCategory:

    def test_value_type(self):
        for val, label in [(1, 1), ("1", 1), (1, "1"), (1.0, 1.0), (1, None)]:
            ac = AttributeCategory(value=val, label=label)
            assert isinstance(ac.value, str)
            if ac.label is not None:
                assert isinstance(ac.label, str)



import logging

import pytest

from arthurai import ArthurAI


BASE_URL = "http://mock"
ACCESS_KEY = "FAKE_ACCESS_KEY"
USER_ACCESS_KEY = "USER_ACCESS_KEY"


@pytest.fixture(scope="class")
def mock_client():
    logging.info("Creating mock ArthurAI Client")

    access_key = "FAKE_ACCESS_KEY"

    config = {
        'access_key': access_key,
        'url': BASE_URL,
        'offline': True
    }
    return ArthurAI(config)


import json
from datetime import datetime
from typing import Dict, Any
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytz
import responses

from tests.fixtures.mocks import BASE_URL


CONTENT_TYPE_JSON = 'application/json'


def mock_post(append_url, dict_to_return, status=200, headers=None):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param dict_to_return: the dictionary to return as json
    :return: None
    """
    if not headers:
        headers = {}

    responses.add(responses.POST, BASE_URL + append_url,
                  json.dumps(dict_to_return), status=status,
                  content_type=CONTENT_TYPE_JSON, 
                  headers = headers)


def mock_patch(append_url, dict_to_return, status=200):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param dict_to_return: the dictionary to return as json
    :return: None
    """

    responses.add(responses.PATCH, BASE_URL + append_url,
                  json.dumps(dict_to_return), status=status,
                  content_type=CONTENT_TYPE_JSON)


def mock_get(append_url, response_body, status=200, content_type=CONTENT_TYPE_JSON):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param response_content: the response_content to return from the mock API request
    :return: None
    """

    responses.add(responses.GET,
                  BASE_URL + append_url,
                  body=json.dumps(response_body),
                  status=status,
                  content_type=content_type)


class MockShortUUID:
    def __init__(self, ids=None):
        if ids is None:
            self.ids = []
        else:
            self.ids = ids
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self):
        value = self.ids[self.next_idx]
        self.next_idx += 1
        return value


class MockDatetime:
    def __init__(self, return_vals=None, return_raw_strings=False):
        if return_vals is None or len(return_vals) == 0:
            self.timestamps = []
        else:
            if isinstance(return_vals[0], datetime):
                self.timestamps = return_vals
            elif isinstance(return_vals[0], str):
                mocks = []
                for ts in return_vals:
                    if return_raw_strings:
                        mocks.append(ts)
                    else:
                        dt = MagicMock(datetime)
                        dt.isoformat.return_value = ts
                        mocks.append(dt)
                self.timestamps = mocks
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self, tz):
        if tz != pytz.utc:
            raise ValueError("unexpected timezone")
        value = self.timestamps[self.next_idx]
        self.next_idx += 1
        return value


def assert_kwargs_equal(actual: Dict[str, Any], expected: Dict[str, Any]):
    assert actual.keys() == expected.keys()
    for key in expected.keys():
        if isinstance(actual[key], (pd.Series, pd.DataFrame)):
            assert actual[key].equals(expected[key])
        elif isinstance(actual[key], np.ndarray):
            assert np.equal(actual[key], expected[key]).all()
        else:
            assert actual[key] == expected[key]



from datetime import datetime, timezone
from mock import patch, mock_open

import numpy as np
import pandas as pd
import pytest
import responses

from arthurai import ArthurAI
from arthurai.common.constants import ValueType, Stage, InputType, OutputType, ImageResponseType, ImageContentType
from arthurai.core.attributes import AttributeCategory, AttributeBin, ArthurAttribute
from arthurai.core.models import ArthurModel
import arthurai.core.util as core_util
from arthurai.common.exceptions import UserValueError, MethodNotApplicableError
from tests.base_test import BaseTest
from tests.test_request_models.fixtures import model_response_json_strings


class TestArthurModel(BaseTest):
    @responses.activate
    def test_send_parquet_file(self):
        input_dict = {
            "attr1": np.Inf,
            "attr2": "string1",
            "attr3": 3.44,
            "attr4": True,
            "attr5": np.nan,
            "attr6": -np.inf,
            "attr7": "string2",
            "attr8": None,
            "attr9": "",
            "attr10": np.inf,
        }
        expected_dict = {
            "attr1": None,
            "attr2": "string1",
            "attr3": 3.44,
            "attr4": True,
            "attr5": None,
            "attr6": None,
            "attr7": "string2",
            "attr8": None,
            "attr9": "",
            "attr10": None,
        }
        output_dict = ArthurModel._replace_nans_and_infinities_in_dict(input_dict)
        assert expected_dict == output_dict

    def test_replace_nans_none_type(self):
        output_dict = ArthurModel._replace_nans_and_infinities_in_dict(None)
        assert output_dict is None

    def test_standardize_pd_obj_for_nans(self):
        input_df = pd.DataFrame(
            [
                [1, 2, 3., 4, 5, 6, np.inf, np.nan, 1, 1],
                [7, 8, 9., 1, 2, 3, np.nan, -np.inf, 2, 2],
                [4, 5.5, 9., -np.inf, np.nan, np.Inf, np.nan, np.nan, np.nan, np.nan],
            ],
            columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
        )
        expected_df = pd.DataFrame(
            [
                [1, 2, 3., 4, 5, 6, np.nan, np.nan, 1, 1],
                [7, 8, 9., 1, 2, 3, np.nan, np.nan, 2, 2],
                [4, 5.5, 9., np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
        )
        expected_df.x3 = expected_df.x3.astype(pd.Int64Dtype())
        expected_df.x4 = expected_df.x4.astype(pd.Int64Dtype())
        expected_df.x5 = expected_df.x5.astype(pd.Int64Dtype())

        attributes = {
            "x3": ValueType.Integer,
            "x4": ValueType.Integer,
            "x5": ValueType.Integer,
            "x8": ValueType.Float,
        }

        assert expected_df.equals(
            core_util.standardize_pd_obj(input_df, dropna=False, replacedatetime=False, attributes=attributes)
        )
        with pytest.raises(ValueError):
            assert expected_df.equals(
                core_util.standardize_pd_obj(input_df, dropna=True, replacedatetime=False, attributes=attributes)
            )  

    def test_standardize_pd_obj_for_series(self):
        input_series = [
            pd.Series([1, 2, 3]),
            pd.Series([1., 2., 3.]),
            pd.Series([1, 2, np.Inf], name="x1"),
            pd.Series([1, 2, np.Inf], name="x1"),
            pd.Series([1, 2, np.Inf], name="x1"),
            pd.Series([np.nan, -np.inf, np.inf]),
        ]
        input_attributes = [
            {},
            None,
            {"x1": ValueType.Integer},
            {"x1": ValueType.Float},
            {"x2": ValueType.Integer},
            {}
        ]
        expected_undropped_series = [
            pd.Series([1, 2, 3]),
            pd.Series([1., 2., 3.]),
            pd.Series([1, 2, np.nan]).astype(pd.Int64Dtype()),
            pd.Series([1, 2, np.nan]),
            pd.Series([1, 2, np.nan]),
            pd.Series([np.nan, np.nan, np.nan]),
        ]
        expected_dropped_series = [
            pd.Series([1, 2, 3]),
            pd.Series([1., 2., 3.]),
            pd.Series([1, 2]).astype(pd.Int64Dtype()),
            pd.Series([1, 2.]),
            pd.Series([1, 2.]),
            pd.Series([]),
        ]

        for inp, inp_attributes, exp_undropped, exp_dropped in zip(input_series, input_attributes, expected_undropped_series, expected_dropped_series):
            print(inp_attributes)
            assert exp_undropped.equals(
                core_util.standardize_pd_obj(inp, dropna=False, replacedatetime=False, attributes=inp_attributes)
            )
            assert exp_dropped.equals(
                core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False, attributes=inp_attributes)
            )

    def test_standardize_pd_obj_bad_input(self):
        for inp in ["string", "", 123, [1,2,3], {"key": "value"}]:
            with pytest.raises(TypeError):
                core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False)

    def test_standardize_pd_obj_for_datetimes(self):
        unaware_timestamp = datetime(2021, 2, 8, 12, 10, 5)
        aware_timestamp = datetime(2021, 2, 8, 12, 10, 5, tzinfo=timezone.utc)
        timestamp_to_expected_output = (
            (unaware_timestamp, ValueError),
            (aware_timestamp, "2021-02-08T12:10:05Z"),
        )

        for timestamp, expected_output in timestamp_to_expected_output:
            if isinstance(expected_output, str):
                actual_output = core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)
                assert expected_output == actual_output.values[0]
            else:
                with pytest.raises(expected_output):
                    core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)

    def test_add_attribute_with_categories(self):
        model = ArthurModel.from_json(model_response_json_strings[2])
        model.add_attribute(
            name="test_categorical_1",
            value_type=ValueType.Integer,
            stage=Stage.ModelPipelineInput,
            categorical=True,
            categories=[1, 2, 3, 4],
        )
        attr = model.get_attribute(name="test_categorical_1")
        assert len(attr.categories) == 4
        assert isinstance(attr.categories[0], AttributeCategory)

        model.add_attribute(
            name="test_categorical_2",
            value_type=ValueType.Integer,
            stage=Stage.ModelPipelineInput,
            categorical=True,
            categories=[AttributeCategory(value="hello")],
        )
        attr = model.get_attribute(name="test_categorical_2")
        assert len(attr.categories) == 1
        assert isinstance(attr.categories[0], AttributeCategory)
        assert attr.categories[0].value == "hello"

    def test_add_attribute_with_bins(self):
        model = ArthurModel.from_json(model_response_json_strings[2])
        model.add_attribute(
            name="test_bins_1",
            value_type=ValueType.Integer,
            stage=Stage.ModelPipelineInput,
            categorical=True,
            bins=[None, 10, 15, 20, None],
        )
        attr = model.get_attribute(name="test_bins_1")
        assert len(attr.bins) == 4
        assert isinstance(attr.bins[0], AttributeBin)
        assert attr.bins[0].continuous_start is None
        assert attr.bins[0].continuous_end == 10

        model.add_attribute(
            name="test_bins_2",
            value_type=ValueType.Integer,
            stage=Stage.ModelPipelineInput,
            categorical=True,
            bins=[
                AttributeBin(continuous_start=None, continuous_end=10),
                AttributeBin(continuous_start=10, continuous_end=15),
                AttributeBin(continuous_start=15, continuous_end=20),
            ],
        )
        attr = model.get_attribute(name="test_bins_2")
        assert len(attr.bins) == 3
        assert isinstance(attr.bins[0], AttributeBin)
        assert attr.bins[0].continuous_start is None
        assert attr.bins[0].continuous_end == 10

    def test_add_nlp_attribute(self):
        model = ArthurModel(
            partner_model_id="test",
            client=None,
            input_type=InputType.NLP,
            output_type=OutputType.Multiclass,
        )
        ref_data = pd.DataFrame(
            {"input_text": ["hello", "hi", "what's up", "yo", "cool"],}
        )
        model.from_dataframe(ref_data, stage=Stage.ModelPipelineInput)

        assert (
            model.get_attribute("input_text").value_type == ValueType.Unstructured_Text
        )
        assert model.get_attribute("input_text").categories is None

    def test_all_null_column(self):
        model = ArthurModel(
            partner_model_id="test",
            client=None,
            input_type=InputType.NLP,
            output_type=OutputType.Multiclass,
        )
        df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
        model.from_dataframe(df, stage=Stage.ModelPipelineInput)

        expected_attributes = [
            ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                            categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
            ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                            categorical=False, is_unique=False)]

        assert model.attributes == expected_attributes

    def test_nonzero_index_column(self):
        model = ArthurModel(
            partner_model_id="test",
            client=None,
            input_type=InputType.NLP,
            output_type=OutputType.Multiclass,
        )
        df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
        df.index = [5]
        model.from_dataframe(df, stage=Stage.ModelPipelineInput)

        expected_attributes = [
            ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                            categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
            ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                            categorical=False, is_unique=False)]

        assert model.attributes == expected_attributes

    def test_set_categorical_value_labels(self):
        model = ArthurModel.from_json(model_response_json_strings[2])
        model.add_attribute(
            name="test_categorical_1",
            value_type=ValueType.Integer,
            stage=Stage.ModelPipelineInput,
            categorical=True,
            categories=[1, 2],
        )
        expected_categories = {1: "Male", 2: "Female"}
        model.set_attribute_labels("test_categorical_1", labels=expected_categories)
        cats = model.get_attribute("test_categorical_1").categories
        for category in cats:
            assert int(category.value) in expected_categories
            assert expected_categories[int(category.value)] == category.label

    @responses.activate
    def test_get_image(self):
        client = ArthurAI(access_key=self.access_key, url=self._base_url, offline=True)
        model = ArthurModel(client=client.client,
                            partner_model_id="test",
                            input_type=InputType.Image,
                            output_type=OutputType.Multiclass)
        model.id = '1234567890abcdef'
        image_id = '0a1b2c3d4e5f9687'
        path = '.'
        type = ImageResponseType.RawImage
        file_ext = '.png'
        response_content = 'content'.encode()
        expected_image_file = f"{path}/{type}_{image_id}{file_ext}"

        self.mockGet(f"/api/v3/models/{model.id}/inferences/images/{image_id}?type={type}", response_content, content_type=ImageContentType.Png)
        open_mock = mock_open()
        with patch("arthurai.core.models.open", open_mock, create=True):
            resulting_file = model.get_image(image_id, path, type=type)

        assert resulting_file == expected_image_file
        open_mock.assert_called_with(expected_image_file, 'wb')
        open_mock.return_value.write.assert_called_once_with(response_content)


    def test_add_object_detection_output_attributes(self):
        model = ArthurModel(
            partner_model_id="test",
            client=None,
            input_type=InputType.Image,
            output_type=OutputType.Multiclass,
        )

        # cannot call without proper types
        with pytest.raises(MethodNotApplicableError):
            model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
        model.input_type = InputType.NLP
        model.output_type = OutputType.ObjectDetection
        with pytest.raises(MethodNotApplicableError):
            model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])

        # cannot call without classes
        model.output_type = OutputType.ObjectDetection
        model.input_type = InputType.Image
        with pytest.raises(UserValueError):
            model.add_object_detection_output_attributes('pred', 'gt', [])
        
        # names cannot match
        with pytest.raises(UserValueError):
            model.add_object_detection_output_attributes('pred', 'pred', ['class_1', 'class_2'])

        # happy path
        model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
        assert model.get_attribute('pred').value_type == ValueType.BoundingBox
        assert model.get_attribute('gt').value_type == ValueType.BoundingBox
        assert model.image_class_labels == ['class_1', 'class_2']